#import <OpenGL/CGLMacro.h>
#import "CTVNC.h"
#import "CTUtil.h"

#include <rfb/rfbclient.h>

///////////////////////////////////////////////////////////////////////////////////

@interface CTWeakViewController : NSViewController
@end

///////////////////////////////////////////////////////////////////////////////////

@interface VNCClient : NSObject
{
    rfbClient *cl;
    BOOL bKeepGoing;

    CTContext *ctcontext;

    NSOpenGLContext *oglctx;
    GLuint framebuffer;
    GLuint fboTex;

    u_short port;
    NSString *host;
    NSString *username;
    NSString *password;
}

- (id)initWithCTContext:(CTContext *)ctx;

- (BOOL)isConnected;

// Private
- (BOOL)resize;
- (rfbCredential *)getCredentials:(int)type;
- (char *)getPassword;
- (void)updateX:(int)x Y:(int)y W:(int)w H:(int)h;

@end

///////////////////////////////////////////////////////////////////////////////////

static int someVar;

static rfbBool CTVNCresize(rfbClient *client)
{
    VNCClient *g = (__bridge VNCClient *)rfbClientGetClientData(client, &someVar);
    return [g resize];
}

static void CTVNCupdate(rfbClient *client, int x, int y, int w, int h)
{
    VNCClient *g = (__bridge VNCClient *)rfbClientGetClientData(client, &someVar);
    [g updateX:x Y:y W:w H:h];
}

static rfbCredential *CTVNCgetcred(rfbClient *client, int type)
{
    VNCClient *g = (__bridge VNCClient *)rfbClientGetClientData(client, &someVar);
    return [g getCredentials:type];
}

static char *CTVNCgetpass(rfbClient *client)
{
    VNCClient *g = (__bridge VNCClient *)rfbClientGetClientData(client, &someVar);
    return [g getPassword];
}

///////////////////////////////////////////////////////////////////////////////////

@implementation VNCClient

- (id)initWithCTContext:(CTContext *)ctx
{
    self = [super init];

    ctcontext = ctx;

    return self;
}

- (void)dealloc
{
    NSLog(@"client dealloc");
}

- (void)_fireConnectedKVO
{
    [self willChangeValueForKey:@"connected"];
    [self didChangeValueForKey:@"connected"];
}

- (void)connectTo:(NSString *)theHost onPort:(u_short)thePort withUserName:(NSString *)user andPassword:(NSString *)pass
{
    host = theHost;
    port = thePort;
    username = user;
    password = pass;

    bKeepGoing = YES;

    [NSThread detachNewThreadSelector:@selector(worker) toTarget:self withObject:nil];
}

- (void)disconnect
{
    @synchronized(self)
    {
        bKeepGoing = NO;
    }
}

- (BOOL)isConnected
{
    @synchronized(self)
    {
        return cl != nil;
    }
}

- (BOOL)resize
{
    // NSLog(@"resize");

    int width = cl->width;
    int height = cl->height;
    int depth = cl->format.bitsPerPixel;

    // NSLog(@"%w:%d h:%d d:%d", width, height, depth);

    if (cl->frameBuffer)
        free(cl->frameBuffer);

    cl->frameBuffer = (uint8_t *)malloc(width * height * depth / 8);

    @synchronized(self)
    {
        NSOpenGLPixelFormatAttribute attributes[] = {NSOpenGLPFANoRecovery, NSOpenGLPFAAccelerated,
                                                     NSOpenGLPFADepthSize, (NSOpenGLPixelFormatAttribute)24,
                                                     (NSOpenGLPixelFormatAttribute)0};

        NSOpenGLPixelFormat *fmt = [[NSOpenGLPixelFormat alloc] initWithAttributes:attributes];

        oglctx = [[NSOpenGLContext alloc] initWithFormat:fmt shareContext:ctcontext.oglCtx];

        CGLContextObj cgl_ctx = [oglctx CGLContextObj];

        {
            glGenFramebuffersEXT(1, &framebuffer);
            glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, framebuffer);

            glGenTextures(1, &fboTex);

            glBindTexture(GL_TEXTURE_RECTANGLE_EXT, fboTex);
            glTexParameteri(GL_TEXTURE_RECTANGLE_EXT, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
            glTexParameteri(GL_TEXTURE_RECTANGLE_EXT, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
            glTexImage2D(GL_TEXTURE_RECTANGLE_EXT, 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
            glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, framebuffer);
            glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_RECTANGLE_EXT, fboTex,
                                      0);
        }

        glViewport(0, 0, width, height);
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        glOrtho(0, width, 0, height, -1, 1);
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

        [CTContext clearContext:oglctx.CGLContextObj toRed:0 green:0 blue:0 alpha:1];
    }

    return true;
}

- (rfbCredential *)getCredentials:(int)type
{
    // NSLog(@"Looking for cred type %d", type);

    if (type == rfbCredentialTypeUser)
    {
        rfbCredential *cred = (rfbCredential *)malloc(sizeof(rfbCredential));
        cred->userCredential.username = strdup(username ? [username UTF8String] : "");
        cred->userCredential.password = strdup(password ? [password UTF8String] : "");
        return cred;
    }

    return nil;
}

- (char *)getPassword
{
    return strdup(password ? [password UTF8String] : "");
}

- (void)updateX:(int)x Y:(int)y W:(int)w H:(int)h
{
#if 0
	x = y = 0;
	w = cl->width;
	h = cl->height;
#endif

    // NSLog(@"update %d %d %d %d", x, y, w, h);

    int width = cl->width;
    int height = cl->height;

    CGLContextObj cgl_ctx = (CGLContextObj)[self->oglctx CGLContextObj];
    CGLSetCurrentContext(cgl_ctx);

    ////

    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glPixelStorei(GL_UNPACK_SWAP_BYTES, 0);
    glPixelStorei(GL_UNPACK_LSB_FIRST, 0);
    glPixelStorei(GL_UNPACK_ROW_LENGTH, width);  // This is the row stride, I think.
    glPixelStorei(GL_UNPACK_IMAGE_HEIGHT, 0);    // This is determined some other way?
    glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
    glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);

    GLuint texture;
    glGenTextures(1, &texture);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, texture);

    // Where is the first pixel in the framebuffer
    unsigned char *pix = cl->frameBuffer;

    // Advance to the first update row
    pix += y * width * 4;

    // Advance to the first pixel in the row
    pix += x * 4;

    // Create a texure with just the updated area
    glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGB, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, pix);

    // Draw it into our scene
    CGRect frect = CGRectMake(0, 0, w, h);
    CGRect trect = CGRectMake(x, y, w, h);

    [CTContext drawTexture:texture fromRect:frect toContext:oglctx.CGLContextObj inRect:trect flipped:YES mirrored:NO blending:NO];

    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
    glDeleteTextures(1, &texture);
}

- (void)worker
{
    @autoreleasepool
    {
        @synchronized(self)
        {
            cl = rfbGetClient(8, 3, 4);

            rfbClientSetClientData(cl, &someVar, (__bridge void *)self);

            cl->MallocFrameBuffer = CTVNCresize;
            cl->canHandleNewFBSize = TRUE;
            cl->GotFrameBufferUpdate = CTVNCupdate;
            // cl->HandleKeyboardLedState=kbd_leds;
            // cl->HandleTextChat=text_chat;
            // cl->GotXCutText = got_selection;
            // cl->listenPort = LISTEN_PORT_OFFSET;
            cl->serverPort = port;
            cl->serverHost = strdup([host UTF8String]);
            cl->GetCredential = CTVNCgetcred;
            cl->GetPassword = CTVNCgetpass;

            // If we have a username, use ARD else use VNC
            uint32 authSchemes = [username length] ? rfbARD : rfbVncAuth;
            SetClientAuthSchemes(cl, &authSchemes, 1);

            if (!rfbInitClient(cl, nil, nil))
            {
                cl = nil;  // rfbInitClient has already freed the client struct
                bKeepGoing = NO;
            }
        }
        // NSLog(@"Connected");

        for (;;)
        {
            @synchronized(self)
            {
                if (!bKeepGoing)
                    break;
            }

            int i = WaitForMessage(cl, 2000000);

            if (i < 0)
                break;

            if (i > 0 && !HandleRFBServerMessage(cl))
                break;
        }

        @synchronized(self)
        {
            if (cl)
                rfbClientCleanup(cl);
            cl = nil;
        }

        [self performSelectorOnMainThread:@selector(_fireConnectedKVO) withObject:nil waitUntilDone:NO];
    }

    NSLog(@"worker exit");
}

- (void)doitToContext:(NSOpenGLContext *)otherContext
{
    CGRect whereRect = [CTUtil fitRect:CGSizeMake(cl->width, cl->height) inContainer:ctcontext.size];

    @synchronized(self)
    {
        if (fboTex)
        {
            [CTContext drawTexture:fboTex
                          fromRect:CGRectMake(0, 0, cl->width, cl->height)
                         toContext:otherContext.CGLContextObj
                            inRect:whereRect
                           flipped:NO
                          mirrored:NO
                          blending:NO];
        }
    }
}

@end

///////////////////////////////////////////////////////////////////////////////////

@implementation CTVNC

+ (NSString *)name
{
    return @"VNC";
}

+ (BOOL)isSource
{
    return YES;
}

- (void)initCommon
{
    viewController =
        [[CTWeakViewController alloc] initWithNibName:@"CTVNC" bundle:[NSBundle bundleForClass:[CTVNC class]]];

    [viewController setRepresentedObject:self];
}

- (id)initWithContext:(CTContext *)ctContext
{
    self = [super initWithContext:ctContext];

    host = @"";
    self.port = 5900;
    username = @"";
    password = @"";
    autoConnect = NO;

    [self initCommon];

    return self;
}

- (void)setConnected:(BOOL)b
{
    // Aways start fresh
    [client disconnect];
    client = nil;

    if (b)
    {
        client = [[VNCClient alloc] initWithCTContext:self.context];
        [client connectTo:host onPort:self.port withUserName:username andPassword:password];
    }
}

- (id)initWithCoder:(NSCoder *)coder
{
    self = [super initWithCoder:coder];

    host = [coder decodeObjectForKey:@"host"];
    self.port = [coder decodeIntForKey:@"port"];
    password = [coder decodeObjectForKey:@"password"];
    username = [coder decodeObjectForKey:@"username"];
    autoConnect = [coder decodeBoolForKey:@"autoConnect"];

    [self initCommon];

    if (autoConnect)
        [self setConnected:YES];

    return self;
}

- (NSView *)inspectorView
{
    return [viewController view];
}

- (void)encodeWithCoder:(NSCoder *)coder
{
    [super encodeWithCoder:coder];

    [coder encodeObject:host forKey:@"host"];
    [coder encodeInt:self.port forKey:@"port"];
    [coder encodeObject:password forKey:@"password"];
    [coder encodeObject:username forKey:@"username"];
    [coder encodeBool:autoConnect forKey:@"autoConnect"];
}

- (void)dealloc
{
    //	NSLog(@"ctvnc dealloc");

    [client disconnect];
}

- (BOOL)isConnected
{
    client && [client isConnected];
}

- (void)doit
{
    [client doitToContext:[[self context] oglCtx]];
}

@end
